@@ -59,7 +59,7 @@
 //! ```toml
 //! [dependencies]
 //! tokio = { version = "1.4", features = ["full"] }
-//! pyo3 = { version = "0.14.1", features = [] }
+//! pyo3 = { version = "0.14.1" }
 //! serde-value = "0.7.0"
 //! pime = "*"
 //! ```
@@ -101,7 +101,7 @@
 //!         engine.launch(&py, broker).unwrap();
 //!     });
 //! });
-//! // wait runtime to be started
+//! // wait engine to be started
 //! pime::wait_online();
 //! 
 //! // Done! Now tasks can be called from any coroutine
@@ -120,7 +120,7 @@
 //!         // The result is returned as Option<Value>
 //!         println!("{:?}", result);
 //!     },
-//!     Err(e) if e.kind == pime::ErrorKind::PythonError => {
+//!     Err(e) if e.kind == pime::ErrorKind::PyException => {
 //!         println!("Exception raised {}: {}", e.exception.unwrap(), e.message);
 //!         println!("{}", e.traceback.unwrap());
 //!     }
@@ -143,6 +143,11 @@
 //!     else:
 //!         raise RuntimeError('command unsupported')
 //! ```
+//! 
+//! ## More examples
+//! 
+//! https://github.com/alttch/pime/tree/main/examples/
+//! 
 use log::{debug, error};
 use pyo3::prelude::*;
 use pythonize::{depythonize, pythonize};
@@ -157,6 +162,8 @@ pub const STATE_STARTING: u8 = 1;
 pub const STATE_STOPPING: u8 = 2;
 pub const STATE_STARTED: u8 = 0xff;
 
+const DATACHANNEL_DEFAULT_BUFFER: usize = 1024;
+
 use std::sync::atomic::{AtomicU8, Ordering};
 
 #[macro_use]
@@ -172,7 +179,7 @@ static ENGINE_STATE: AtomicU8 = AtomicU8::new(STATE_STOPPED);
 
 #[derive(Debug, Eq, PartialEq)]
 pub enum ErrorKind {
-    PythonError,
+    PyException,
     PackError,
     UnpackError,
     ExecError,
@@ -220,7 +227,7 @@ impl Error {
     }
     fn new_py(error: (String, String, String)) -> Self {
         Self {
-            kind: ErrorKind::PythonError,
+            kind: ErrorKind::PyException,
             exception: Some(error.0),
             message: error.1,
             traceback: Some(error.2),
@@ -260,6 +267,7 @@ pub struct PyTask {
     command: Value,
     params: BTreeMap<String, Value>,
     need_result: bool,
+    exclusive: bool,
 }
 
 impl PyTask {
@@ -268,6 +276,7 @@ impl PyTask {
             command,
             params,
             need_result: true,
+            exclusive: false,
         })
     }
 
@@ -276,12 +285,18 @@ impl PyTask {
             command,
             params: BTreeMap::new(),
             need_result: true,
+            exclusive: false,
         })
     }
 
     pub fn no_wait(&mut self) {
         self.need_result = false;
     }
+
+    pub fn mark_exclusive(&mut self) {
+        self.exclusive = true;
+        self.need_result = true;
+    }
 }
 
 struct DataChannel {
@@ -291,12 +306,18 @@ struct DataChannel {
 
 impl DataChannel {
     fn new() -> Self {
-        let (tx, rx) = mpsc::channel::<(u64, Option<Box<PyTask>>)>(1024);
+        let (tx, rx) = mpsc::channel::<(u64, Option<Box<PyTask>>)>(DATACHANNEL_DEFAULT_BUFFER);
         Self {
             tx: Mutex::new(tx),
             rx: Mutex::new(rx),
         }
     }
+
+    fn set_buffer(&mut self, buffer: usize) {
+        let (tx, rx) = mpsc::channel::<(u64, Option<Box<PyTask>>)>(buffer);
+        self.tx = Mutex::new(tx);
+        self.rx = Mutex::new(rx);
+    }
 }
 
 #[derive(Debug)]
@@ -341,7 +362,7 @@ lazy_static! {
     static ref PY_RESULTS: RwLock<Box<BTreeMap<u64, PyTaskResult>>> =
         RwLock::new(Box::new(BTreeMap::new()));
     static ref TASK_COUNTER: Mutex<PyTaskCounter> = Mutex::new(PyTaskCounter::new());
-    static ref DC: DataChannel = DataChannel::new();
+    static ref DC: RwLock<DataChannel> = RwLock::new(DataChannel::new());
 }
 
 pub struct PySyncEngine<'p> {
@@ -454,7 +475,7 @@ impl<'p> PySyncEngine<'p> {
     pub fn add_import_path(&self, path: &str) -> Result<(), Error> {
         match self.neo.call_method1("add_import_path", (path,)) {
             Ok(_) => Ok(()),
-            Err(e) => Err(Error::new(ErrorKind::PythonError, tostr!(e))),
+            Err(e) => Err(Error::new(ErrorKind::PyException, tostr!(e))),
         }
     }
 
@@ -477,10 +498,12 @@ impl<'p> PySyncEngine<'p> {
     pub fn launch(&self, py: &'p pyo3::Python, broker: &pyo3::PyAny) -> Result<(), Error> {
         need_offline!();
 
-        let mut rx = DC.rx.try_lock()?;
+        let dc = DC.try_read()?;
+        let mut rx = dc.rx.try_lock()?;
 
         self.neo.call_method0("start")?;
         let call = self.neo.getattr("call")?;
+        let call_direct = self.neo.getattr("call_direct")?;
         let spawn = self.neo.getattr("spawn")?;
 
         ENGINE_STATE.store(STATE_STARTED, Ordering::SeqCst);
@@ -518,7 +541,19 @@ impl<'p> PySyncEngine<'p> {
                         }
                         continue;
                     }
-                    if task.need_result {
+                    if task.exclusive {
+                        match call_direct.call1((
+                            task_id,
+                            broker,
+                            command.unwrap(),
+                            params.unwrap(),
+                        )) {
+                            Ok(_) => {}
+                            Err(e) => {
+                                report_error(task_id, Error::new(ErrorKind::ExecError, tostr!(e)));
+                            }
+                        }
+                    } else if task.need_result {
                         match call.call1((task_id, broker, command.unwrap(), params.unwrap())) {
                             Ok(_) => {}
                             Err(e) => {
@@ -542,7 +577,13 @@ impl<'p> PySyncEngine<'p> {
 pub async fn call(task: Box<PyTask>) -> Result<Option<Box<Value>>, Error> {
     need_online!();
     if !task.need_result {
-        DC.tx.lock().await.send((0, Some(task))).await?;
+        DC.read()
+            .await
+            .tx
+            .lock()
+            .await
+            .send((0, Some(task)))
+            .await?;
         return Ok(None);
     }
     let task_id;
@@ -566,7 +607,13 @@ pub async fn call(task: Box<PyTask>) -> Result<Option<Box<Value>>, Error> {
             ready: beacon.clone(),
         },
     );
-    DC.tx.lock().await.send((task_id, Some(task))).await?;
+    DC.read()
+        .await
+        .tx
+        .lock()
+        .await
+        .send((task_id, Some(task)))
+        .await?;
     beacon.notified().await;
     let res = match PY_RESULTS.write().await.remove(&task_id) {
         Some(v) => v,
@@ -585,7 +632,7 @@ pub async fn call(task: Box<PyTask>) -> Result<Option<Box<Value>>, Error> {
 
 pub async fn stop() -> Result<(), Error> {
     need_online!();
-    DC.tx.lock().await.send((0, None)).await?;
+    DC.read().await.tx.lock().await.send((0, None)).await?;
     wait_offline();
     Ok(())
 }
@@ -609,3 +656,7 @@ pub fn wait_offline() {
         std::thread::sleep(Duration::from_millis(10));
     }
 }
+
+pub fn set_mpsc_buffer(buffer: usize) {
+    DC.try_write().unwrap().set_buffer(buffer);
+}
